import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

category_name = "Test 0470 Tool dependency repository type"
category_description = "Test script 0470 for changing repository types."
package_libx11_repository_name = "package_x11_client_1_5_proto_7_0_0470"
package_libx11_repository_description = (
    "Contains a tool dependency definition that provides the X11 client libraries and core protocol header files."
)
package_libx11_repository_long_description = (
    "Xlib is an X Window System protocol client library written in the C programming language."
)
package_emboss_repository_name = "package_emboss_5_0_0_0470"
package_emboss_repository_description = (
    "Contains a tool dependency definition that downloads and compiles version 5.0.0 of the EMBOSS tool suite."
)
package_emboss_repository_long_description = 'EMBOSS is "The European Molecular Biology Open Software Suite".'
datatypes_repository_name = "emboss_datatypes_0470"
datatypes_repository_description = "Galaxy applicable data formats used by Emboss tools."
datatypes_repository_long_description = (
    "Galaxy applicable data formats used by Emboss tools.  This repository contains no tools."
)
emboss_repository_name = "emboss_5_0470"
emboss_repository_description = "Galaxy wrappers for Emboss version 5.0.0 tools"
emboss_repository_long_description = "Galaxy wrappers for Emboss version 5.0.0 tools"

"""
1. Create and populate a repository named package_x11_client_1_5_proto_7_0 that contains only a single file named tool_dependencies.xml.
   Keep the repository type as the default "Unrestricted".

2. Create a repository named package_emboss_5_0_0 of type "Unrestricted" that has a repository dependency definition that defines the
   above package_x11_client_1_5_proto_7_0 repository. Upload the tool_dependencies.xml file such that it does not have a changeset_revision
   defined so it will get automatically populated.

3. Create a repository named emboss_5 of type "Unrestricted" that has a tool-dependencies.xml file defining a complex repository dependency
   on the package_emboss_5_0_0 repository above. Upload the tool_dependencies.xml file such that it does not have a change set_revision defined
   so it will get automatically populated.

4. Add a comment to the tool_dependencies.xml file to be uploaded to the package_x11_client_1_5_prot_7_0 repository, creating a new installable
   changeset revision at the repository tip.

5. Add a comment to the tool_dependencies.xml file for the package_emboss_5_0_0 repository, eliminating the change set_revision attribute so
   that it gets automatically populated when uploaded. After uploading the file, the package_emboss_5_0_0 repository should have 2
   installable changeset revisions.

6. Add a comment to the tool_dependencies.xml file in the emboss_5 repository, eliminating the changeset_revision attribute so that it gets
   automatically populated when uploaded. After uploading the file, the emboss5 repository should have 2 installable metadata revisions.

7. Change the repository type of the package_x11_client_1_5_proto_7_0 repository to be tool_dependency_definition.

8. Change the repository type of the package_emboss_5_0_0 repository to be tool_dependency_definition.

9. Reset metadata on the package_emboss_5_0_0 repository. It should now have only its tip as the installable revision.

10. Reset metadata on the emboss_5 repository. It should now have only its tip as the installable revision.
"""


class TestEnvironmentInheritance(ShedTwillTestCase):
    """Test referencing environment variables that were defined in a separate tool dependency."""

    def test_0000_initiate_users_and_category(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.admin_email, username=common.admin_username)
        self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)

    def test_0005_create_libx11_repository(self):
        """Create and populate package_x11_client_1_5_proto_7_0_0470.

        This is step 1 - Create and populate a repository named package_x11_client_1_5_proto_7_0.

        Create and populate a repository named package_x11_client_1_5_proto_7_0 that contains only a single file named tool_dependencies.xml.
        Keep the repository type as the default "Unrestricted".
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=package_libx11_repository_name,
            description=package_libx11_repository_description,
            long_description=package_libx11_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        # Upload the tool dependency definition to the package_x11_client_1_5_proto_7_0_0470 repository.
        self.user_populator().setup_test_data_repo("libx11_proto", repository, end=1)

    def test_0010_create_emboss_5_0_0_repository(self):
        """Create and populate package_emboss_5_0_0_0470.

        This is step 2 - Create a repository named package_emboss_5_0_0 of type "Unrestricted".

        Create a repository named package_emboss_5_0_0 of type "Unrestricted" that has a repository dependency definition that defines the
        above package_x11_client_1_5_proto_7_0 repository. Upload the tool_dependencues.xml file such that it does not have a changeset_revision
        defined so it will get automatically populated.
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=package_emboss_repository_name,
            description=package_emboss_repository_description,
            long_description=package_emboss_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        # Upload the edited tool dependency definition to the package_emboss_5_0_0 repository.
        self.user_populator().setup_test_data_repo("package_emboss_5_0_0_0470", repository, end=1)

    def test_0015_create_emboss_5_repository(self):
        """Create and populate emboss_5_0470.

        This is step 3 - Create a repository named emboss_5 of type "Unrestricted".

        Create a repository named emboss_5 of type "Unrestricted" that has a tool-dependencies.xml file defining a complex repository dependency
        on the package_emboss_5_0_0 repository above. Upload the tool_dependencies.xml file such that it does not have a change set_revision defined
        so it will get automatically populated.
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=emboss_repository_name,
            description=emboss_repository_description,
            long_description=emboss_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        # Populate emboss_5 with tool and dependency definitions.
        self.user_populator().setup_test_data_repo("emboss_5_0470", repository, end=1)

    def test_0020_upload_updated_tool_dependency_to_package_x11(self):
        """Upload a new tool_dependencies.xml to package_x11_client_1_5_proto_7_0_0470.

        This is step 4 - Add a comment to the tool_dependencies.xml file to be uploaded to the package_x11_client_1_5_prot_7_0 repository, creating
        a new installable changeset revision at the repository tip.
        """
        package_x11_repository = self._get_repository_by_name_and_owner(
            package_libx11_repository_name, common.test_user_1_name
        )
        # Upload the tool dependency definition to the package_x11_client_1_5_proto_7_0_0470 repository.
        self.user_populator().setup_test_data_repo("libx11_proto", package_x11_repository, start=1, end=2)
        count = self._get_metadata_revision_count(package_x11_repository)
        assert (
            count == 1
        ), f"package_x11_client_1_5_proto_7_0_0470 has incorrect number of metadata revisions, expected 1 but found {count}"

    def test_0025_upload_updated_tool_dependency_to_package_emboss(self):
        """Upload a new tool_dependencies.xml to package_emboss_5_0_0_0470.

        This is step 5 - Add a comment to the tool_dependencies.xml file for the package_emboss_5_0_0 repository, eliminating
        the change set_revision attribute so that it gets automatically populated when uploaded. After uploading the file,
        the package_emboss_5_0_0 repository should have 2 installable changeset revisions.
        """
        package_emboss_repository = self._get_repository_by_name_and_owner(
            package_emboss_repository_name, common.test_user_1_name
        )
        # Populate package_emboss_5_0_0_0470 with updated tool dependency definition.
        self.user_populator().setup_test_data_repo(
            "package_emboss_5_0_0_0470", package_emboss_repository, start=1, end=2
        )
        count = self._get_metadata_revision_count(package_emboss_repository)
        assert (
            count == 2
        ), f"package_emboss_5_0_0_0470 has incorrect number of metadata revisions, expected 2 but found {count}"

    def test_0030_upload_updated_tool_dependency_to_emboss_5_repository(self):
        """Upload a new tool_dependencies.xml to emboss_5_0470.

        This is step 6 - Add a comment to the tool_dependencies.xml file in the emboss_5 repository, eliminating the
        changeset_revision attribute so that it gets automatically populated when uploaded. After uploading the file,
        the emboss5 repository should have 2 installable metadata revisions.
        """
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        # Populate package_emboss_5_0_0_0470 with updated tool dependency definition.
        self.user_populator().setup_test_data_repo("emboss_5_0470", emboss_repository, start=1, end=2)
        count = self._get_metadata_revision_count(emboss_repository)
        assert count == 2, "package_emboss_5_0_0_0470 has incorrect number of metadata revisions"

    def test_0035_modify_package_x11_repository_type(self):
        """Set package_x11_client_1_5_proto_7_0 type tool_dependency_definition.

        This is step 7 - Change the repository type of the package_x11_client_1_5_proto_7_0 repository to be tool_dependency_definition.
        """
        package_x11_repository = self._get_repository_by_name_and_owner(
            package_libx11_repository_name, common.test_user_1_name
        )
        self.edit_repository_information(package_x11_repository, repository_type="tool_dependency_definition")

    def test_0040_modify_package_emboss_repository_type(self):
        """Set package_emboss_5_0_0 to type tool_dependency_definition.

        This is step 8 - Change the repository type of the package_emboss_5_0_0 repository to be tool_dependency_definition.
        """
        package_emboss_repository = self._get_repository_by_name_and_owner(
            package_emboss_repository_name, common.test_user_1_name
        )
        self.edit_repository_information(package_emboss_repository, repository_type="tool_dependency_definition")

    def test_0045_reset_repository_metadata(self):
        """Reset metadata on package_emboss_5_0_0_0470 and package_x11_client_1_5_proto_7_0.

        This is step 9 - Reset metadata on the package_emboss_5_0_0 and package_x11_client_1_5_proto_7_0 repositories. They should
        now have only their tip as the installable revision.
        """
        package_emboss_repository = self._get_repository_by_name_and_owner(
            package_emboss_repository_name, common.test_user_1_name
        )
        package_x11_repository = self._get_repository_by_name_and_owner(
            package_libx11_repository_name, common.test_user_1_name
        )
        self.reset_repository_metadata(package_emboss_repository)
        self.reset_repository_metadata(package_x11_repository)
        count = self._get_metadata_revision_count(package_emboss_repository)
        assert count == 1, f"Repository package_emboss_5_0_0 has {count} installable revisions, expected 1."
        count = self._get_metadata_revision_count(package_x11_repository)
        assert count == 1, f"Repository package_x11_client_1_5_proto_7_0 has {count} installable revisions, expected 1."

    def test_0050_reset_emboss_5_metadata(self):
        """Reset metadata on emboss_5.

        This is step 10 - Reset metadata on the emboss_5 repository. It should now have only its tip as the installable revision.
        """
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        self.reset_repository_metadata(emboss_repository)
        count = self._get_metadata_revision_count(emboss_repository)
        assert count == 1, f"Repository emboss_5 has {count} installable revisions, expected 1."



import json
import logging

from galaxy import (
    exceptions,
    util,
)
from galaxy.web import (
    expose_api,
    expose_api_raw_anonymous_and_sessionless,
    require_admin,
)
from tool_shed.managers.tools import search
from tool_shed.util.shed_index import build_index
from . import BaseShedAPIController

log = logging.getLogger(__name__)


class ToolsController(BaseShedAPIController):
    """RESTful controller for interactions with tools in the Tool Shed."""

    @expose_api
    @require_admin
    def build_search_index(self, trans, **kwd):
        """
        PUT /api/tools/build_search_index

        Not part of the stable API, just something to simplify bootstrapping tool sheds,
        scripting, testing, etc...
        """
        repos_indexed, tools_indexed = build_index(
            trans.app.config.whoosh_index_dir,
            trans.app.config.file_path,
            trans.app.config.hgweb_config_dir,
            trans.app.config.hgweb_repo_prefix,
            trans.app.config.database_connection,
        )
        return {
            "repositories_indexed": repos_indexed,
            "tools_indexed": tools_indexed,
        }

    @expose_api_raw_anonymous_and_sessionless
    def index(self, trans, **kwd):
        """
        GET /api/tools
        Displays a collection of tools with optional criteria.

        :param q:        (optional)if present search on the given query will be performed
        :type  q:        str

        :param page:     (optional)requested page of the search
        :type  page:     int

        :param page_size:     (optional)requested page_size of the search
        :type  page_size:     int

        :param jsonp:    (optional)flag whether to use jsonp format response, defaults to False
        :type  jsonp:    bool

        :param callback: (optional)name of the function to wrap callback in
                         used only when jsonp is true, defaults to 'callback'
        :type  callback: str

        :returns dict:   object containing list of results and metadata

        Examples:
            GET http://localhost:9009/api/tools
            GET http://localhost:9009/api/tools?q=fastq
        """
        q = kwd.get("q", "")
        if not q:
            raise exceptions.NotImplemented(
                'Listing of all the tools is not implemented. Provide parameter "q" to search instead.'
            )
        else:
            page = kwd.get("page", 1)
            page_size = kwd.get("page_size", 10)
            try:
                page = int(page)
                page_size = int(page_size)
            except ValueError:
                raise exceptions.RequestParameterInvalidException('The "page" and "page_size" have to be integers.')
            return_jsonp = util.asbool(kwd.get("jsonp", False))
            callback = kwd.get("callback", "callback")
            search_results = search(trans, q, page, page_size)
            if return_jsonp:
                response = str(f"{callback}({json.dumps(search_results)});")
            else:
                response = json.dumps(search_results)
            return response



"""Tool Shed Security"""

import logging
from typing import List

from sqlalchemy import (
    false,
    select,
)

from galaxy.util import listify
from galaxy.util.bunch import Bunch
from tool_shed.webapp.model import (
    Group,
    Role,
)

IUC_NAME = "Intergalactic Utilities Commission"

log = logging.getLogger(__name__)


class Action:
    def __init__(self, action, description, model):
        self.action = action
        self.description = description
        self.model = model


class RBACAgent:
    """Handle Galaxy Tool Shed security"""

    permitted_actions = Bunch()

    def associate_components(self, **kwd):
        raise Exception(f"No valid method of associating provided components: {kwd}")

    def associate_user_role(self, user, role):
        raise Exception("No valid method of associating a user with a role")

    def convert_permitted_action_strings(self, permitted_action_strings):
        """
        When getting permitted actions from an untrusted source like a
        form, ensure that they match our actual permitted actions.
        """
        return [
            x
            for x in [self.permitted_actions.get(action_string) for action_string in permitted_action_strings]
            if x is not None
        ]

    def create_user_role(self, user, app):
        raise Exception("Unimplemented Method")

    def create_private_user_role(self, user):
        raise Exception("Unimplemented Method")

    def get_action(self, name, default=None):
        """Get a permitted action by its dict key or action name"""
        for k, v in self.permitted_actions.items():
            if k == name or v.action == name:
                return v
        return default

    def get_actions(self):
        """Get all permitted actions as a list of Action objects"""
        return list(self.permitted_actions.__dict__.values())

    def get_item_actions(self, action, item):
        raise Exception(f"No valid method of retrieving action ({action}) for item {item}.")

    def get_private_user_role(self, user):
        raise Exception("Unimplemented Method")


class CommunityRBACAgent(RBACAgent):
    def __init__(self, model, permitted_actions=None):
        self.model = model
        if permitted_actions:
            self.permitted_actions = permitted_actions

    @property
    def sa_session(self):
        """Returns a SQLAlchemy scoped session"""
        return self.model.context

    def allow_action(self, roles, action, item):
        """
        Method for checking a permission for the current user ( based on roles ) to perform a
        specific action on an item
        """
        item_actions = self.get_item_actions(action, item)
        if not item_actions:
            return action.model == "restrict"
        ret_val = False
        for item_action in item_actions:
            if item_action.role in roles:
                ret_val = True
                break
        return ret_val

    def associate_components(self, **kwd):
        if "user" in kwd:
            if "group" in kwd:
                return self.associate_user_group(kwd["user"], kwd["group"])
            elif "role" in kwd:
                return self.associate_user_role(kwd["user"], kwd["role"])
        elif "role" in kwd:
            if "group" in kwd:
                return self.associate_group_role(kwd["group"], kwd["role"])
        elif "repository" in kwd:
            return self.associate_repository_category(kwd["repository"], kwd["category"])
        raise Exception(f"No valid method of associating provided components: {kwd}")

    def associate_group_role(self, group, role):
        assoc = self.model.GroupRoleAssociation(group, role)
        self.sa_session.add(assoc)
        session = self.sa_session()
        session.commit()
        return assoc

    def associate_user_group(self, user, group):
        assoc = self.model.UserGroupAssociation(user, group)
        self.sa_session.add(assoc)
        session = self.sa_session()
        session.commit()
        return assoc

    def associate_user_role(self, user, role):
        assoc = self.model.UserRoleAssociation(user, role)
        self.sa_session.add(assoc)
        session = self.sa_session()
        session.commit()
        return assoc

    def associate_repository_category(self, repository, category):
        assoc = self.model.RepositoryCategoryAssociation(repository, category)
        self.sa_session.add(assoc)
        session = self.sa_session()
        session.commit()
        return assoc

    def create_private_user_role(self, user):
        # Create private role
        role = self.model.Role(
            name=user.email, description=f"Private Role for {user.email}", type=self.model.Role.types.PRIVATE
        )
        self.sa_session.add(role)
        session = self.sa_session()
        session.commit()
        # Add user to role
        self.associate_components(role=role, user=user)
        return role

    def create_user_role(self, user, app):
        self.get_private_user_role(user, auto_create=True)

    def get_item_actions(self, action, item):
        # item must be one of: Dataset, Library, LibraryFolder, LibraryDataset, LibraryDatasetDatasetAssociation
        return [permission for permission in item.actions if permission.action == action.action]

    def get_private_user_role(self, user, auto_create=False):
        role = _get_private_user_role(self.sa_session, user.email)
        if not role:
            if auto_create:
                return self.create_private_user_role(user)
            else:
                return None
        return role

    def set_entity_group_associations(self, groups=None, users=None, roles=None, delete_existing_assocs=True):
        if groups is None:
            groups = []
        if users is None:
            users = []
        if roles is None:
            roles = []
        for group in groups:
            if delete_existing_assocs:
                for a in group.roles + group.users:
                    self.sa_session.delete(a)
                    session = self.sa_session()
                    session.commit()
            for role in roles:
                self.associate_components(group=group, role=role)
            for user in users:
                self.associate_components(group=group, user=user)

    def set_entity_role_associations(
        self, roles=None, users=None, groups=None, repositories=None, delete_existing_assocs=True
    ):
        if roles is None:
            roles = []
        if users is None:
            users = []
        if groups is None:
            groups = []
        if repositories is None:
            repositories = []
        for role in roles:
            if delete_existing_assocs:
                for a in role.users + role.groups:
                    self.sa_session.delete(a)
                    session = self.sa_session()
                    session.commit()
            for user in users:
                self.associate_components(user=user, role=role)
            for group in groups:
                self.associate_components(group=group, role=role)

    def set_entity_user_associations(self, users=None, roles=None, groups=None, delete_existing_assocs=True):
        if users is None:
            users = []
        if roles is None:
            roles = []
        if groups is None:
            groups = []
        for user in users:
            if delete_existing_assocs:
                for a in user.non_private_roles + user.groups:
                    self.sa_session.delete(a)
                    session = self.sa_session()
                    session.commit()
            self.sa_session.refresh(user)
            for role in roles:
                # Make sure we are not creating an additional association with a PRIVATE role
                if role not in user.roles:
                    self.associate_components(user=user, role=role)
            for group in groups:
                self.associate_components(user=user, group=group)

    def usernames_that_can_push(self, repository) -> List[str]:
        return listify(repository.allow_push())

    def can_push(self, app, user, repository):
        if user:
            return user.username in self.usernames_that_can_push(repository)
        return False

    def user_can_administer_repository(self, user, repository):
        """Return True if the received user can administer the received repository."""
        if user:
            if repository:
                repository_admin_role = repository.admin_role
                for rra in repository.roles:
                    role = rra.role
                    if role.id == repository_admin_role.id:
                        # We have the repository's admin role, so see if the user is associated with it.
                        for ura in role.users:
                            role_member = ura.user
                            if role_member.id == user.id:
                                return True
                        # The user is not directly associated with the role, so see if they are a member
                        # of a group that is associated with the role.
                        for gra in role.groups:
                            group = gra.group
                            for uga in group.users:
                                member = uga.user
                                if member.id == user.id:
                                    return True
        return False

    def user_can_import_repository_archive(self, user, archive_owner):
        # This method should be called only if the current user is not an admin.
        if user.username == archive_owner:
            return True
        # A member of the IUC is authorized to create new repositories that are owned by another user.
        if (iuc_group := get_iuc_group(self.sa_session)) is not None:
            for uga in iuc_group.users:
                if uga.user.id == user.id:
                    return True
        return False


def get_permitted_actions(filter=None):
    """Utility method to return a subset of RBACAgent's permitted actions"""
    if filter is None:
        return RBACAgent.permitted_actions
    tmp_bunch = Bunch()
    [tmp_bunch.__dict__.__setitem__(k, v) for k, v in RBACAgent.permitted_actions.items() if k.startswith(filter)]
    return tmp_bunch


def get_iuc_group(session):
    stmt = select(Group).where(Group.name == IUC_NAME).where(Group.deleted == false()).limit(1)
    return session.scalars(stmt).first()


def _get_private_user_role(session, user_email):
    stmt = select(Role).where(Role.name == user_email).where(Role.type == Role.types.PRIVATE).limit(1)
    return session.scalars(stmt).first()



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

freebayes_repository_name = "freebayes_0040"
freebayes_repository_description = "Galaxy's freebayes tool for test 0040"
freebayes_repository_long_description = "Long description of Galaxy's freebayes tool for test 0040"

filtering_repository_name = "filtering_0040"
filtering_repository_description = "Galaxy's filtering tool for test 0040"
filtering_repository_long_description = "Long description of Galaxy's filtering tool for test 0040"

category_name = "test_0040_repository_circular_dependencies"

running_standalone = False


class TestInstallingCircularDependencies(ShedTwillTestCase):
    """Verify that the code correctly handles installing repositories with circular dependencies."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_freebayes_repository(self):
        """Create and populate freebayes_0040."""
        global running_standalone
        category = self.create_category(
            name=category_name, description="Testing handling of circular repository dependencies."
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=freebayes_repository_name,
            description=freebayes_repository_description,
            long_description=freebayes_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            running_standalone = True
            self.commit_tar_to_repository(
                repository,
                "freebayes/freebayes.tar",
                commit_message="Uploaded the tool tarball.",
            )

    def test_0015_create_filtering_repository(self):
        """Create and populate filtering_0040."""
        global running_standalone
        category = self.create_category(
            name=category_name, description="Testing handling of circular repository dependencies."
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=filtering_repository_name,
            description=filtering_repository_description,
            long_description=filtering_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            running_standalone = True
            self.commit_tar_to_repository(
                repository,
                "filtering/filtering_1.1.0.tar",
                commit_message="Uploaded the tool tarball for filtering 1.1.0.",
            )

    def test_0020_create_repository_dependencies(self):
        """Set up the filtering and freebayes repository dependencies."""
        # The dependency structure should look like:
        # Filtering revision 0 -> freebayes revision 0.
        # Freebayes revision 0 -> filtering revision 1.
        # Filtering will have two revisions, one with just the filtering tool, and one with the filtering tool and a dependency on freebayes.
        if running_standalone:
            freebayes_repository = self._get_repository_by_name_and_owner(
                freebayes_repository_name, common.test_user_1_name
            )
            filtering_repository = self._get_repository_by_name_and_owner(
                filtering_repository_name, common.test_user_1_name
            )
            repository_dependencies_path = self.generate_temp_path("test_1040", additional_paths=["circular"])
            repository_tuple = (
                self.url,
                freebayes_repository.name,
                freebayes_repository.owner,
                self.get_repository_tip(freebayes_repository),
            )
            self.create_repository_dependency(
                repository=filtering_repository,
                repository_tuples=[repository_tuple],
                filepath=repository_dependencies_path,
            )
            repository_tuple = (
                self.url,
                filtering_repository.name,
                filtering_repository.owner,
                self.get_repository_tip(filtering_repository),
            )
            self.create_repository_dependency(
                repository=freebayes_repository,
                repository_tuples=[repository_tuple],
                filepath=repository_dependencies_path,
            )

    def test_0025_install_freebayes_repository(self):
        """Install freebayes with blank tool panel section, without tool dependencies but with repository dependencies."""
        self._install_repository(
            freebayes_repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
            install_repository_dependencies=True,
        )

    def test_0030_uninstall_filtering_repository(self):
        """Deactivate filtering, verify tool panel section and missing repository dependency."""
        installed_freebayes_repository = self._get_installed_repository_by_name_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        installed_filtering_repository = self._get_installed_repository_by_name_owner(
            filtering_repository_name, common.test_user_1_name
        )
        assert self._get_installed_repository_for(
            common.test_user_1, freebayes_repository_name, installed_freebayes_repository.installed_changeset_revision
        )
        assert self._get_installed_repository_for(
            common.test_user_1, filtering_repository_name, installed_filtering_repository.installed_changeset_revision
        )
        self.deactivate_repository(installed_filtering_repository)
        self._refresh_tool_shed_repository(installed_filtering_repository)
        self._assert_has_missing_dependency(installed_freebayes_repository, filtering_repository_name)
        self.check_galaxy_repository_db_status(filtering_repository_name, common.test_user_1_name, "Deactivated")

    def test_0035_reactivate_filtering_repository(self):
        """Reinstall filtering into 'filtering' tool panel section."""
        installed_filtering_repository = self._get_installed_repository_by_name_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        self.reinstall_repository_api(
            installed_filtering_repository,
            install_tool_dependencies=False,
            install_repository_dependencies=True,
            new_tool_panel_section_label="filtering",
        )
        installed_freebayes_repository = self._get_installed_repository_by_name_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        self._assert_is_not_missing_dependency(installed_freebayes_repository, filtering_repository_name)

    def test_0040_uninstall_freebayes_repository(self):
        """Deactivate freebayes, verify tool panel section and missing repository dependency."""
        installed_freebayes_repository = self._get_installed_repository_by_name_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        installed_filtering_repository = self._get_installed_repository_by_name_owner(
            filtering_repository_name, common.test_user_1_name
        )
        assert self._get_installed_repository_for(
            common.test_user_1, freebayes_repository_name, installed_freebayes_repository.installed_changeset_revision
        )
        assert self._get_installed_repository_for(
            common.test_user_1, filtering_repository_name, installed_filtering_repository.installed_changeset_revision
        )
        self.deactivate_repository(installed_freebayes_repository)
        assert not self._get_installed_repository_for(
            common.test_user_1, freebayes_repository_name, installed_freebayes_repository.installed_changeset_revision
        )
        self._refresh_tool_shed_repository(installed_filtering_repository)
        self._assert_has_missing_dependency(installed_filtering_repository, freebayes_repository_name)
        self.check_galaxy_repository_db_status("freebayes_0040", "user1", "Deactivated")

    def test_0045_deactivate_filtering_repository(self):
        """Deactivate filtering, verify tool panel section."""
        installed_filtering_repository = self._get_installed_repository_by_name_owner(
            filtering_repository_name, common.test_user_1_name
        )
        installed_freebayes_repository = self._get_installed_repository_by_name_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        assert self._get_installed_repository_for(
            common.test_user_1, filtering_repository_name, installed_filtering_repository.installed_changeset_revision
        )
        self.deactivate_repository(installed_filtering_repository)
        assert not self._get_installed_repository_for(
            common.test_user_1, freebayes_repository_name, installed_freebayes_repository.installed_changeset_revision
        )
        assert not self._get_installed_repository_for(
            common.test_user_1, filtering_repository_name, installed_filtering_repository.installed_changeset_revision
        )
        self._refresh_tool_shed_repository(installed_freebayes_repository)
        self._assert_has_missing_dependency(installed_freebayes_repository, filtering_repository_name)
        self.check_galaxy_repository_db_status(filtering_repository_name, common.test_user_1_name, "Deactivated")



import logging
import time
from typing import Optional

from sqlalchemy import (
    false,
    func,
    select,
)

from galaxy import (
    util,
    web,
)
from galaxy.model import (
    Library,
    LibraryDatasetDatasetAssociation,
)
from galaxy.security.validate_user_input import validate_password
from galaxy.util import inflector
from galaxy.util.hash_util import new_secure_hash_v2
from galaxy.web.form_builder import CheckboxField
from galaxy.web.legacy_framework.grids import (
    Grid,
    GridOperation,
)
from tool_shed.util.web_util import escape
from tool_shed.webapp import model

log = logging.getLogger(__name__)
compliance_log = logging.getLogger("COMPLIANCE")


class Admin:
    # Override these
    user_list_grid: Optional[Grid] = None
    role_list_grid: Optional[Grid] = None
    group_list_grid: Optional[Grid] = None
    delete_operation: Optional[GridOperation] = None
    undelete_operation: Optional[GridOperation] = None
    purge_operation: Optional[GridOperation] = None

    @web.expose
    @web.require_admin
    def index(self, trans, **kwd):
        message = escape(kwd.get("message", ""))
        status = kwd.get("status", "done")
        return trans.fill_template("/webapps/tool_shed/admin/index.mako", message=message, status=status)

    @web.expose
    @web.require_admin
    def center(self, trans, **kwd):
        message = escape(kwd.get("message", ""))
        status = kwd.get("status", "done")
        return trans.fill_template("/webapps/tool_shed/admin/center.mako", message=message, status=status)

    @web.expose
    @web.require_admin
    def roles(self, trans, **kwargs):
        if "operation" in kwargs:
            operation = kwargs["operation"].lower().replace("+", " ")
            if operation == "roles":
                return self.role(trans, **kwargs)
            if operation == "create":
                return self.create_role(trans, **kwargs)
            if operation == "delete":
                return self.mark_role_deleted(trans, **kwargs)
            if operation == "undelete":
                return self.undelete_role(trans, **kwargs)
            if operation == "purge":
                return self.purge_role(trans, **kwargs)
            if operation == "manage users and groups":
                return self.manage_users_and_groups_for_role(trans, **kwargs)
            if operation == "manage role associations":
                # This is currently used only in the Tool Shed.
                return self.manage_role_associations(trans, **kwargs)
            if operation == "rename":
                return self.rename_role(trans, **kwargs)
        # Render the list view
        return self.role_list_grid(trans, **kwargs)

    @web.expose
    @web.require_admin
    def create_role(self, trans, **kwd):
        params = util.Params(kwd)
        message = util.restore_text(params.get("message", ""))
        status = params.get("status", "done")
        name = util.restore_text(params.get("name", ""))
        description = util.restore_text(params.get("description", ""))
        in_users = util.listify(params.get("in_users", []))
        out_users = util.listify(params.get("out_users", []))
        in_groups = util.listify(params.get("in_groups", []))
        out_groups = util.listify(params.get("out_groups", []))
        create_group_for_role = params.get("create_group_for_role", "")
        create_group_for_role_checked = CheckboxField.is_checked(create_group_for_role)
        ok = True

        if params.get("create_role_button", False):
            if not name or not description:
                message = "Enter a valid name and a description."
                status = "error"
                ok = False
            elif get_role_id(trans.sa_session, model.Role, name):
                message = "Role names must be unique and a role with that name already exists, so choose another name."
                status = "error"
                ok = False
            else:
                # Create the role
                role = model.Role(name=name, description=description, type=model.Role.types.ADMIN)
                trans.sa_session.add(role)
                # Create the UserRoleAssociations
                for user in [trans.sa_session.get(model.User, x) for x in in_users]:
                    ura = model.UserRoleAssociation(user, role)
                    trans.sa_session.add(ura)
                # Create the GroupRoleAssociations
                for group in [trans.sa_session.get(model.Group, x) for x in in_groups]:
                    gra = model.GroupRoleAssociation(group, role)
                    trans.sa_session.add(gra)
                if create_group_for_role_checked:
                    # Create the group
                    group = model.Group(name=name)
                    trans.sa_session.add(group)
                    # Associate the group with the role
                    gra = model.GroupRoleAssociation(group, role)
                    trans.sa_session.add(gra)
                    num_in_groups = len(in_groups) + 1
                else:
                    num_in_groups = len(in_groups)
                trans.sa_session.commit()
                message = f"Role '{role.name}' has been created with {len(in_users)} associated users and {num_in_groups} associated groups.  "
                if create_group_for_role_checked:
                    message += (
                        "One of the groups associated with this role is the newly created group with the same name."
                    )
                trans.response.send_redirect(
                    web.url_for(controller="admin", action="roles", message=util.sanitize_text(message), status="done")
                )
        if ok:
            for user in get_current_users(trans.sa_session, model.User):
                out_users.append((user.id, user.email))
            for group in get_current_groups(trans.sa_session, model.Group):
                out_groups.append((group.id, group.name))
        return trans.fill_template(
            "/webapps/tool_shed/admin/dataset_security/role/role_create.mako",
            name=name,
            description=description,
            in_users=in_users,
            out_users=out_users,
            in_groups=in_groups,
            out_groups=out_groups,
            create_group_for_role_checked=create_group_for_role_checked,
            message=message,
            status=status,
        )

    @web.expose
    @web.require_admin
    def rename_role(self, trans, **kwd):
        params = util.Params(kwd)
        message = util.restore_text(params.get("message", ""))
        status = params.get("status", "done")
        id = params.get("id", None)
        if not id:
            message = "No role ids received for renaming"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="roles", message=message, status="error")
            )
        role = get_role(trans, id)
        if params.get("rename_role_button", False):
            old_name = role.name
            new_name = util.restore_text(params.name)
            new_description = util.restore_text(params.description)
            if not new_name:
                message = "Enter a valid name"
                status = "error"
            else:
                if get_role_id(trans.sa_session, model.Role, new_name) != role.id:
                    message = "A role with that name already exists"
                    status = "error"
                else:
                    if not (role.name == new_name and role.description == new_description):
                        role.name = new_name
                        role.description = new_description
                        trans.sa_session.add(role)
                        trans.sa_session.commit()
                        message = f"Role '{old_name}' has been renamed to '{new_name}'"
                    return trans.response.send_redirect(
                        web.url_for(
                            controller="admin", action="roles", message=util.sanitize_text(message), status="done"
                        )
                    )
        return trans.fill_template(
            "/webapps/tool_shed/admin/dataset_security/role/role_rename.mako", role=role, message=message, status=status
        )

    @web.expose
    @web.require_admin
    def manage_users_and_groups_for_role(self, trans, **kwd):
        params = util.Params(kwd)
        message = util.restore_text(params.get("message", ""))
        status = params.get("status", "done")
        id = params.get("id", None)
        if not id:
            message = "No role ids received for managing users and groups"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="roles", message=message, status="error")
            )
        role = get_role(trans, id)
        if params.get("role_members_edit_button", False):
            in_users = [trans.sa_session.get(model.User, x) for x in util.listify(params.in_users)]
            if trans.webapp.name == "galaxy":
                for ura in role.users:
                    user = trans.sa_session.get(model.User, ura.user_id)
                    if user not in in_users:
                        # Delete DefaultUserPermissions for previously associated users that have been removed from the role
                        for dup in user.default_permissions:
                            if role == dup.role:
                                trans.sa_session.delete(dup)
                        # Delete DefaultHistoryPermissions for previously associated users that have been removed from the role
                        for history in user.histories:
                            for dhp in history.default_permissions:
                                if role == dhp.role:
                                    trans.sa_session.delete(dhp)
                        trans.sa_session.commit()
            in_groups = [trans.sa_session.get(model.Group, x) for x in util.listify(params.in_groups)]
            trans.app.security_agent.set_entity_role_associations(roles=[role], users=in_users, groups=in_groups)
            trans.sa_session.refresh(role)
            message = f"Role '{role.name}' has been updated with {len(in_users)} associated users and {len(in_groups)} associated groups"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="roles", message=util.sanitize_text(message), status=status)
            )
        in_users = []
        out_users = []
        in_groups = []
        out_groups = []
        for user in get_current_users(trans.sa_session, model.User):
            if user in [x.user for x in role.users]:
                in_users.append((user.id, user.email))
            else:
                out_users.append((user.id, user.email))
        for group in get_current_groups(trans.sa_session, model.Group):
            if group in [x.group for x in role.groups]:
                in_groups.append((group.id, group.name))
            else:
                out_groups.append((group.id, group.name))
        library_dataset_actions = {}
        if trans.webapp.name == "galaxy" and len(role.dataset_actions) < 25:
            # Build a list of tuples that are LibraryDatasetDatasetAssociationss followed by a list of actions
            # whose DatasetPermissions is associated with the Role
            # [ ( LibraryDatasetDatasetAssociation [ action, action ] ) ]
            for dp in role.dataset_actions:
                for ldda in get_ldda_by_dataset(trans.sa_session, dp.dataset_id):
                    root_found = False
                    folder_path = ""
                    folder = ldda.library_dataset.folder
                    while not root_found:
                        folder_path = f"{folder.name} / {folder_path}"
                        if not folder.parent:
                            root_found = True
                        else:
                            folder = folder.parent
                    folder_path = f"{folder_path} {ldda.name}"
                    library = get_library_by_folder(trans.sa_session, folder.id)
                    if library not in library_dataset_actions:
                        library_dataset_actions[library] = {}
                    try:
                        library_dataset_actions[library][folder_path].append(dp.action)
                    except Exception:
                        library_dataset_actions[library][folder_path] = [dp.action]
        else:
            message = "Not showing associated datasets, there are too many."
            status = "info"
        return trans.fill_template(
            "/webapps/tool_shed/admin/dataset_security/role/role.mako",
            role=role,
            in_users=in_users,
            out_users=out_users,
            in_groups=in_groups,
            out_groups=out_groups,
            library_dataset_actions=library_dataset_actions,
            message=message,
            status=status,
        )

    @web.expose
    @web.require_admin
    def mark_role_deleted(self, trans, **kwd):
        id = kwd.get("id", None)
        if not id:
            message = "No role ids received for deleting"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="roles", message=message, status="error")
            )
        ids = util.listify(id)
        message = f"Deleted {len(ids)} roles: "
        for role_id in ids:
            role = get_role(trans, role_id)
            role.deleted = True
            trans.sa_session.add(role)
            trans.sa_session.commit()
            message += f" {role.name} "
        trans.response.send_redirect(
            web.url_for(controller="admin", action="roles", message=util.sanitize_text(message), status="done")
        )

    @web.expose
    @web.require_admin
    def undelete_role(self, trans, **kwd):
        id = kwd.get("id", None)
        if not id:
            message = "No role ids received for undeleting"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="roles", message=message, status="error")
            )
        ids = util.listify(id)
        undeleted_roles = []
        for role_id in ids:
            role = get_role(trans, role_id)
            if not role.deleted:
                message = f"Role '{role.name}' has not been deleted, so it cannot be undeleted."
                trans.response.send_redirect(
                    web.url_for(controller="admin", action="roles", message=util.sanitize_text(message), status="error")
                )
            role.deleted = False
            trans.sa_session.add(role)
            trans.sa_session.commit()
            undeleted_roles.append(role.name)
        message = "Undeleted {} roles: {}".format(len(undeleted_roles), " ".join(undeleted_roles))
        trans.response.send_redirect(
            web.url_for(controller="admin", action="roles", message=util.sanitize_text(message), status="done")
        )

    @web.expose
    @web.require_admin
    def purge_role(self, trans, **kwd):
        # This method should only be called for a Role that has previously been deleted.
        # Purging a deleted Role deletes all of the following from the database:
        # - UserRoleAssociations where role_id == Role.id
        # - DefaultUserPermissions where role_id == Role.id
        # - DefaultHistoryPermissions where role_id == Role.id
        # - GroupRoleAssociations where role_id == Role.id
        # - DatasetPermissionss where role_id == Role.id
        id = kwd.get("id", None)
        if not id:
            message = "No role ids received for purging"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="roles", message=util.sanitize_text(message), status="error")
            )
        ids = util.listify(id)
        message = f"Purged {len(ids)} roles: "
        for role_id in ids:
            role = get_role(trans, role_id)
            if not role.deleted:
                message = f"Role '{role.name}' has not been deleted, so it cannot be purged."
                trans.response.send_redirect(
                    web.url_for(controller="admin", action="roles", message=util.sanitize_text(message), status="error")
                )
            # Delete UserRoleAssociations
            for ura in role.users:
                user = trans.sa_session.get(model.User, ura.user_id)
                # Delete DefaultUserPermissions for associated users
                for dup in user.default_permissions:
                    if role == dup.role:
                        trans.sa_session.delete(dup)
                # Delete DefaultHistoryPermissions for associated users
                for history in user.histories:
                    for dhp in history.default_permissions:
                        if role == dhp.role:
                            trans.sa_session.delete(dhp)
                trans.sa_session.delete(ura)
            # Delete GroupRoleAssociations
            for gra in role.groups:
                trans.sa_session.delete(gra)
            # Delete DatasetPermissionss
            for dp in role.dataset_actions:
                trans.sa_session.delete(dp)
            trans.sa_session.commit()
            message += f" {role.name} "
        trans.response.send_redirect(
            web.url_for(controller="admin", action="roles", message=util.sanitize_text(message), status="done")
        )

    @web.expose
    @web.require_admin
    def groups(self, trans, **kwargs):
        if "operation" in kwargs:
            operation = kwargs["operation"].lower().replace("+", " ")
            if operation == "groups":
                return self.group(trans, **kwargs)
            if operation == "create":
                return self.create_group(trans, **kwargs)
            if operation == "delete":
                return self.mark_group_deleted(trans, **kwargs)
            if operation == "undelete":
                return self.undelete_group(trans, **kwargs)
            if operation == "purge":
                return self.purge_group(trans, **kwargs)
            if operation == "manage users and roles":
                return self.manage_users_and_roles_for_group(trans, **kwargs)
            if operation == "rename":
                return self.rename_group(trans, **kwargs)
        # Render the list view
        return self.group_list_grid(trans, **kwargs)

    @web.expose
    @web.require_admin
    def rename_group(self, trans, **kwd):
        params = util.Params(kwd)
        message = util.restore_text(params.get("message", ""))
        status = params.get("status", "done")
        id = params.get("id", None)
        if not id:
            message = "No group ids received for renaming"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="groups", message=message, status="error")
            )
        group = get_group(trans, id)
        if params.get("rename_group_button", False):
            old_name = group.name
            new_name = util.restore_text(params.name)
            if not new_name:
                message = "Enter a valid name"
                status = "error"
            else:
                if get_group_id(trans.sa_session, model.Group, new_name) != group.id:
                    message = "A group with that name already exists"
                    status = "error"
                else:
                    if group.name != new_name:
                        group.name = new_name
                        trans.sa_session.add(group)
                        trans.sa_session.commit()
                        message = f"Group '{old_name}' has been renamed to '{new_name}'"
                    return trans.response.send_redirect(
                        web.url_for(
                            controller="admin", action="groups", message=util.sanitize_text(message), status="done"
                        )
                    )
        return trans.fill_template(
            "/webapps/tool_shed/admin/dataset_security/group/group_rename.mako",
            group=group,
            message=message,
            status=status,
        )

    @web.expose
    @web.require_admin
    def manage_users_and_roles_for_group(self, trans, **kwd):
        params = util.Params(kwd)
        message = util.restore_text(params.get("message", ""))
        status = params.get("status", "done")
        group = get_group(trans, params.id)
        if params.get("group_roles_users_edit_button", False):
            in_roles = [trans.sa_session.get(model.Role, x) for x in util.listify(params.in_roles)]
            in_users = [trans.sa_session.get(model.User, x) for x in util.listify(params.in_users)]
            trans.app.security_agent.set_entity_group_associations(groups=[group], roles=in_roles, users=in_users)
            trans.sa_session.refresh(group)
            message += f"Group '{group.name}' has been updated with {len(in_roles)} associated roles and {len(in_users)} associated users"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="groups", message=util.sanitize_text(message), status=status)
            )
        in_roles = []
        out_roles = []
        in_users = []
        out_users = []
        for role in get_current_roles(trans.sa_session, model.Role):
            if role in [x.role for x in group.roles]:
                in_roles.append((role.id, role.name))
            else:
                out_roles.append((role.id, role.name))
        for user in get_current_users(trans.sa_session, model.User):
            if user in [x.user for x in group.users]:
                in_users.append((user.id, user.email))
            else:
                out_users.append((user.id, user.email))
        message += f"Group {group.name} is currently associated with {len(in_roles)} roles and {len(in_users)} users"
        return trans.fill_template(
            "/webapps/tool_shed/admin/dataset_security/group/group.mako",
            group=group,
            in_roles=in_roles,
            out_roles=out_roles,
            in_users=in_users,
            out_users=out_users,
            message=message,
            status=status,
        )

    @web.expose
    @web.require_admin
    def create_group(self, trans, **kwd):
        params = util.Params(kwd)
        message = util.restore_text(params.get("message", ""))
        status = params.get("status", "done")
        name = util.restore_text(params.get("name", ""))
        in_users = util.listify(params.get("in_users", []))
        out_users = util.listify(params.get("out_users", []))
        in_roles = util.listify(params.get("in_roles", []))
        out_roles = util.listify(params.get("out_roles", []))
        create_role_for_group = params.get("create_role_for_group", "")
        create_role_for_group_checked = CheckboxField.is_checked(create_role_for_group)
        ok = True
        if params.get("create_group_button", False):
            if not name:
                message = "Enter a valid name."
                status = "error"
                ok = False
            elif get_group_id(trans.sa_session, model.Group, name):
                message = (
                    "Group names must be unique and a group with that name already exists, so choose another name."
                )
                status = "error"
                ok = False
            else:
                # Create the group
                group = model.Group(name=name)
                trans.sa_session.add(group)
                trans.sa_session.commit()
                # Create the UserRoleAssociations
                for user in [trans.sa_session.get(model.User, x) for x in in_users]:
                    uga = model.UserGroupAssociation(user, group)
                    trans.sa_session.add(uga)
                # Create the GroupRoleAssociations
                for role in [trans.sa_session.get(model.Role, x) for x in in_roles]:
                    gra = model.GroupRoleAssociation(group, role)
                    trans.sa_session.add(gra)
                if create_role_for_group_checked:
                    # Create the role
                    role = model.Role(name=name, description=f"Role for group {name}")
                    trans.sa_session.add(role)
                    # Associate the role with the group
                    gra = model.GroupRoleAssociation(group, role)
                    trans.sa_session.add(gra)
                    num_in_roles = len(in_roles) + 1
                else:
                    num_in_roles = len(in_roles)
                trans.sa_session.commit()
                message = f"Group '{group.name}' has been created with {len(in_users)} associated users and {num_in_roles} associated roles.  "
                if create_role_for_group_checked:
                    message += (
                        "One of the roles associated with this group is the newly created role with the same name."
                    )
                trans.response.send_redirect(
                    web.url_for(controller="admin", action="groups", message=util.sanitize_text(message), status="done")
                )
        if ok:
            for user in get_current_users(trans.sa_session, model.User):
                out_users.append((user.id, user.email))
            for role in get_current_roles(trans.sa_session, model.Role):
                out_roles.append((role.id, role.name))
        return trans.fill_template(
            "/webapps/tool_shed/admin/dataset_security/group/group_create.mako",
            name=name,
            in_users=in_users,
            out_users=out_users,
            in_roles=in_roles,
            out_roles=out_roles,
            create_role_for_group_checked=create_role_for_group_checked,
            message=message,
            status=status,
        )

    @web.expose
    @web.require_admin
    def mark_group_deleted(self, trans, **kwd):
        params = util.Params(kwd)
        id = params.get("id", None)
        if not id:
            message = "No group ids received for marking deleted"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="groups", message=message, status="error")
            )
        ids = util.listify(id)
        message = f"Deleted {len(ids)} groups: "
        for group_id in ids:
            group = get_group(trans, group_id)
            group.deleted = True
            trans.sa_session.add(group)
            trans.sa_session.commit()
            message += f" {group.name} "
        trans.response.send_redirect(
            web.url_for(controller="admin", action="groups", message=util.sanitize_text(message), status="done")
        )

    @web.expose
    @web.require_admin
    def undelete_group(self, trans, **kwd):
        id = kwd.get("id", None)
        if not id:
            message = "No group ids received for undeleting"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="groups", message=message, status="error")
            )
        ids = util.listify(id)
        undeleted_groups = []
        for group_id in ids:
            group = get_group(trans, group_id)
            if not group.deleted:
                message = f"Group '{group.name}' has not been deleted, so it cannot be undeleted."
                trans.response.send_redirect(
                    web.url_for(
                        controller="admin", action="groups", message=util.sanitize_text(message), status="error"
                    )
                )
            group.deleted = False
            trans.sa_session.add(group)
            trans.sa_session.commit()
            undeleted_groups.append(group.name)
        message = "Undeleted {} groups: {}".format(len(undeleted_groups), " ".join(undeleted_groups))
        trans.response.send_redirect(
            web.url_for(controller="admin", action="groups", message=util.sanitize_text(message), status="done")
        )

    @web.expose
    @web.require_admin
    def purge_group(self, trans, **kwd):
        # This method should only be called for a Group that has previously been deleted.
        # Purging a deleted Group simply deletes all UserGroupAssociations and GroupRoleAssociations.
        id = kwd.get("id", None)
        if not id:
            message = "No group ids received for purging"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="groups", message=util.sanitize_text(message), status="error")
            )
        ids = util.listify(id)
        message = f"Purged {len(ids)} groups: "
        for group_id in ids:
            group = get_group(trans, group_id)
            if not group.deleted:
                # We should never reach here, but just in case there is a bug somewhere...
                message = f"Group '{group.name}' has not been deleted, so it cannot be purged."
                trans.response.send_redirect(
                    web.url_for(
                        controller="admin", action="groups", message=util.sanitize_text(message), status="error"
                    )
                )
            # Delete UserGroupAssociations
            for uga in group.users:
                trans.sa_session.delete(uga)
            # Delete GroupRoleAssociations
            for gra in group.roles:
                trans.sa_session.delete(gra)
            trans.sa_session.commit()
            message += f" {group.name} "
        trans.response.send_redirect(
            web.url_for(controller="admin", action="groups", message=util.sanitize_text(message), status="done")
        )

    @web.expose
    @web.require_admin
    def create_new_user(self, trans, **kwd):
        return trans.response.send_redirect(web.url_for(controller="user", action="create", cntrller="admin"))

    @web.expose
    @web.require_admin
    def reset_user_password(self, trans, **kwd):
        user_id = kwd.get("id", None)
        if not user_id:
            message = "No users received for resetting passwords."
            trans.response.send_redirect(
                web.url_for(controller="admin", action="users", message=message, status="error")
            )
        user_ids = util.listify(user_id)
        if "reset_user_password_button" in kwd:
            message = ""
            status = ""
            for user_id in user_ids:
                user = get_user(trans, user_id)
                password = kwd.get("password", None)
                confirm = kwd.get("confirm", None)
                message = validate_password(trans, password, confirm)
                if message:
                    status = "error"
                    break
                else:
                    user.set_password_cleartext(password)
                    trans.sa_session.add(user)
                    trans.sa_session.commit()
            if not message and not status:
                message = "Passwords reset for {} {}.".format(
                    len(user_ids), inflector.cond_plural(len(user_ids), "user")
                )
                status = "done"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="users", message=util.sanitize_text(message), status=status)
            )
        users = [get_user(trans, user_id) for user_id in user_ids]
        if len(user_ids) > 1:
            user_id = ",".join(user_ids)
        return trans.fill_template(
            "/webapps/tool_shed/admin/user/reset_password.mako", id=user_id, users=users, password="", confirm=""
        )

    @web.expose
    @web.require_admin
    def mark_user_deleted(self, trans, **kwd):
        id = kwd.get("id", None)
        if not id:
            message = "No user ids received for deleting"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="users", message=message, status="error")
            )
        ids = util.listify(id)
        message = f"Deleted {len(ids)} users: "
        for user_id in ids:
            user = get_user(trans, user_id)
            user.deleted = True

            compliance_log.info(f"delete-user-event: {user_id}")
            # See lib/galaxy/webapps/tool_shed/controllers/admin.py
            pseudorandom_value = str(int(time.time()))
            email_hash = new_secure_hash_v2(user.email + pseudorandom_value)
            uname_hash = new_secure_hash_v2(user.username + pseudorandom_value)
            for role in user.all_roles():
                print(
                    role, self.app.config.redact_username_during_deletion, self.app.config.redact_email_during_deletion
                )
                if self.app.config.redact_username_during_deletion:
                    role.name = role.name.replace(user.username, uname_hash)
                    role.description = role.description.replace(user.username, uname_hash)

                if self.app.config.redact_email_during_deletion:
                    role.name = role.name.replace(user.email, email_hash)
                    role.description = role.description.replace(user.email, email_hash)

            if self.app.config.redact_email_during_deletion:
                user.email = email_hash
            if self.app.config.redact_username_during_deletion:
                user.username = uname_hash

            trans.sa_session.add(user)
            trans.sa_session.commit()
            message += f" {user.email} "
        trans.response.send_redirect(
            web.url_for(controller="admin", action="users", message=util.sanitize_text(message), status="done")
        )

    @web.expose
    @web.require_admin
    def undelete_user(self, trans, **kwd):
        id = kwd.get("id", None)
        if not id:
            message = "No user ids received for undeleting"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="users", message=message, status="error")
            )
        ids = util.listify(id)
        undeleted_users = []
        for user_id in ids:
            user = get_user(trans, user_id)
            if not user.deleted:
                message = f"User '{user.email}' has not been deleted, so it cannot be undeleted."
                trans.response.send_redirect(
                    web.url_for(controller="admin", action="users", message=util.sanitize_text(message), status="error")
                )
            user.deleted = False
            trans.sa_session.add(user)
            trans.sa_session.commit()
            undeleted_users.append(user.email)
        message = "Undeleted {} users: {}".format(len(undeleted_users), " ".join(undeleted_users))
        trans.response.send_redirect(
            web.url_for(controller="admin", action="users", message=util.sanitize_text(message), status="done")
        )

    @web.expose
    @web.require_admin
    def purge_user(self, trans, **kwd) -> None:
        # This method should only be called for a User that has previously been deleted.
        # We keep the User in the database ( marked as purged ), and stuff associated
        # with the user's private role in case we want the ability to unpurge the user
        # some time in the future.
        # Purging a deleted User deletes all of the following:
        # - History where user_id = User.id
        #    - HistoryDatasetAssociation where history_id = History.id
        #    - Dataset where HistoryDatasetAssociation.dataset_id = Dataset.id
        # - UserGroupAssociation where user_id == User.id
        # - UserRoleAssociation where user_id == User.id EXCEPT FOR THE PRIVATE ROLE
        # - UserAddress where user_id == User.id
        id = kwd.get("id", None)
        if not id:
            message = "No user ids received for purging"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="users", message=util.sanitize_text(message), status="error")
            )
        ids = util.listify(id)
        message = f"Purged {len(ids)} users: "
        for user_id in ids:
            user = get_user(trans, user_id)
            if not user.deleted:
                # We should never reach here, but just in case there is a bug somewhere...
                message = f"User '{user.email}' has not been deleted, so it cannot be purged."
                trans.response.send_redirect(
                    web.url_for(controller="admin", action="users", message=util.sanitize_text(message), status="error")
                )
            private_role = trans.app.security_agent.get_private_user_role(user)
            # Delete UserGroupAssociations
            for uga in user.groups:
                trans.sa_session.delete(uga)
            # Delete UserRoleAssociations EXCEPT FOR THE PRIVATE ROLE
            for ura in user.roles:
                if ura.role_id != private_role.id:
                    trans.sa_session.delete(ura)
            # Purge the user
            user.purged = True
            trans.sa_session.add(user)
            trans.sa_session.commit()
            message += f"{user.email} "
        trans.response.send_redirect(
            web.url_for(controller="admin", action="users", message=util.sanitize_text(message), status="done")
        )

    @web.expose
    @web.require_admin
    def users(self, trans, **kwd):
        if "operation" in kwd:
            operation = kwd["operation"].lower()
            if operation == "roles":
                return self.user(trans, **kwd)
            elif operation == "reset password":
                return self.reset_user_password(trans, **kwd)
            elif operation == "delete":
                return self.mark_user_deleted(trans, **kwd)
            elif operation == "undelete":
                return self.undelete_user(trans, **kwd)
            elif operation == "purge":
                return self.purge_user(trans, **kwd)
            elif operation == "create":
                return self.create_new_user(trans, **kwd)
            elif operation == "manage roles and groups":
                return self.manage_roles_and_groups_for_user(trans, **kwd)

        if trans.app.config.allow_user_deletion:
            if self.delete_operation not in self.user_list_grid.operations:
                self.user_list_grid.operations.append(self.delete_operation)
            if self.undelete_operation not in self.user_list_grid.operations:
                self.user_list_grid.operations.append(self.undelete_operation)
            if self.purge_operation not in self.user_list_grid.operations:
                self.user_list_grid.operations.append(self.purge_operation)
        # Render the list view
        return self.user_list_grid(trans, **kwd)

    @web.expose
    @web.require_admin
    def name_autocomplete_data(self, trans, q=None, limit=None, timestamp=None):
        """Return autocomplete data for user emails"""
        emails = get_user_emails_by_prefix(trans.sa_session, model.User, q)
        return "\n".join(emails)

    @web.expose
    @web.require_admin
    def manage_roles_and_groups_for_user(self, trans, **kwd):
        user_id = kwd.get("id", None)
        message = ""
        status = ""
        if not user_id:
            message += f"Invalid user id ({str(user_id)}) received"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="users", message=util.sanitize_text(message), status="error")
            )
        user = get_user(trans, user_id)
        private_role = trans.app.security_agent.get_private_user_role(user)
        if kwd.get("user_roles_groups_edit_button", False):
            # Make sure the user is not dis-associating himself from his private role
            out_roles = kwd.get("out_roles", [])
            if out_roles:
                out_roles = [trans.sa_session.get(model.Role, x) for x in util.listify(out_roles)]
            if private_role in out_roles:
                message += "You cannot eliminate a user's private role association.  "
                status = "error"
            in_roles = kwd.get("in_roles", [])
            if in_roles:
                in_roles = [trans.sa_session.get(model.Role, x) for x in util.listify(in_roles)]
            out_groups = kwd.get("out_groups", [])
            if out_groups:
                out_groups = [trans.sa_session.get(model.Group, x) for x in util.listify(out_groups)]
            in_groups = kwd.get("in_groups", [])
            if in_groups:
                in_groups = [trans.sa_session.get(model.Group, x) for x in util.listify(in_groups)]
            if in_roles:
                trans.app.security_agent.set_entity_user_associations(users=[user], roles=in_roles, groups=in_groups)
                trans.sa_session.refresh(user)
                message += f"User '{user.email}' has been updated with {len(in_roles)} associated roles and {len(in_groups)} associated groups (private roles are not displayed)"
                trans.response.send_redirect(
                    web.url_for(controller="admin", action="users", message=util.sanitize_text(message), status="done")
                )
        in_roles = []
        out_roles = []
        in_groups = []
        out_groups = []
        for role in get_current_roles(trans.sa_session, model.Role):
            if role in [x.role for x in user.roles]:
                in_roles.append((role.id, role.name))
            elif role.type != model.Role.types.PRIVATE:
                # There is a 1 to 1 mapping between a user and a PRIVATE role, so private roles should
                # not be listed in the roles form fields, except for the currently selected user's private
                # role, which should always be in in_roles.  The check above is added as an additional
                # precaution, since for a period of time we were including private roles in the form fields.
                out_roles.append((role.id, role.name))
        for group in get_current_groups(trans.sa_session, model.Group):
            if group in [x.group for x in user.groups]:
                in_groups.append((group.id, group.name))
            else:
                out_groups.append((group.id, group.name))
        message += f"User '{user.email}' is currently associated with {len(in_roles)} roles and is a member of {len(in_groups)} groups"
        if not status:
            status = "done"
        return trans.fill_template(
            "/webapps/tool_shed/admin/user/user.mako",
            user=user,
            in_roles=in_roles,
            out_roles=out_roles,
            in_groups=in_groups,
            out_groups=out_groups,
            message=message,
            status=status,
        )


# ---- Utility methods -------------------------------------------------------


def get_user(trans, user_id) -> model.User:
    """Get a User from the database by id."""
    user = trans.sa_session.get(model.User, trans.security.decode_id(user_id))
    if not user:
        return trans.show_error_message(f"User not found for id ({str(user_id)})")
    return user


def get_role(trans, id):
    """Get a Role from the database by id."""
    # Load user from database
    id = trans.security.decode_id(id)
    role = trans.sa_session.get(model.Role, id)
    if not role:
        return trans.show_error_message(f"Role not found for id ({str(id)})")
    return role


def get_group(trans, id):
    """Get a Group from the database by id."""
    # Load user from database
    id = trans.security.decode_id(id)
    group = trans.sa_session.get(model.Group, id)
    if not group:
        return trans.show_error_message(f"Group not found for id ({str(id)})")
    return group


def get_role_id(session, role_model, name):
    stmt = select(role_model.id).where(role_model.name == name).limit(1)
    return session.scalars(stmt).first()


def get_group_id(session, group_model, name):
    stmt = select(group_model.id).where(group_model.name == name).limit(1)
    return session.scalars(stmt).first()


def get_current_users(session, user_model):
    stmt = select(user_model).where(user_model.deleted == false()).order_by(user_model.email)
    return session.scalars(stmt)


def get_current_groups(session, group_model):
    stmt = select(group_model).where(group_model.deleted == false()).order_by(group_model.name)
    return session.scalars(stmt)


def get_current_roles(session, role_model):
    stmt = select(role_model).where(role_model.deleted == false()).order_by(role_model.name)
    return session.scalars(stmt)


def get_ldda_by_dataset(session, dataset_id):
    stmt = select(LibraryDatasetDatasetAssociation).where(LibraryDatasetDatasetAssociation.dataset_id == dataset_id)
    return session.scalars(stmt)


def get_library_by_folder(session, folder_id):
    stmt = select(Library).where(Library.folder_id == folder_id).limit(1)
    return session.scalars(stmt).first()


def get_user_emails_by_prefix(session, user_model, prefix):
    stmt = (
        select(user_model.email)
        .where(user_model.deleted == false())
        .where(func.lower(user_model.email).like(f"{prefix.lower()}%"))
    )
    return session.scalars(stmt)



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

category_name = "Test 1430 Repair installed repository"
category_description = "Test script 1430 for repairing an installed repository."
filter_repository_name = "filter_1430"
column_repository_name = "column_1430"
filter_repository_description = "Galaxy's filter tool for test 1430"
column_repository_description = "Add a value as a new column"
filter_repository_long_description = f"{filter_repository_name}: {filter_repository_description}"
column_repository_long_description = f"{column_repository_name}: {column_repository_description}"

"""
In the Tool Shed:

1) Create and populate the filter_1430 repository

2) Create and populate the column_1430 repository

3) Upload a repository_dependencies.xml file to the column_1430 repository that creates a repository dependency on the filter_1430 repository.

In Galaxy:

1) Install the column_1430 repository, making sure to check the checkbox to Handle repository dependencies so that the filter
   repository is also installed. Make sure to install the repositories in a specified section of the tool panel.

2) Uninstall the filter_1430 repository.
"""


class TestRepairRepository(ShedTwillTestCase):
    """Test repairing an installed repository."""

    requires_galaxy = True

    def test_0000_initiate_users_and_category(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.admin_email, username=common.admin_username)
        self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)

    def test_0005_create_filter_repository(self):
        """Create and populate the filter_1430 repository.

        This is step 1 - Create and populate the filter_1430 repository.

        This repository will be depended on by the column_1430 repository.
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=filter_repository_name,
            description=filter_repository_description,
            long_description=filter_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "filtering/filtering_1.1.0.tar",
            commit_message="Populate filter_1430 with version 1.1.0.",
        )

    def test_0010_create_column_repository(self):
        """Create and populate the column_1430 repository.

        This is step 2 - Create and populate the column_1430 repository.

        This repository will depend on the filter_1430 repository.
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=column_repository_name,
            description=column_repository_description,
            long_description=column_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "column_maker/column_maker.tar",
            commit_message="Populate column_1430 with tool definitions.",
        )

    def test_0015_create_repository_dependency(self):
        """Create a dependency on filter_1430.

        This is step 3 - Upload a repository_dependencies.xml file to the column_1430 repository that creates a repository
        dependency on the filter_1430 repository.
        """
        column_repository = self._get_repository_by_name_and_owner("column_1430", common.test_user_1_name)
        filter_repository = self._get_repository_by_name_and_owner("filter_1430", common.test_user_1_name)
        tool_shed_url = self.url
        name = filter_repository.name
        owner = filter_repository.owner
        changeset_revision = self.get_repository_tip(filter_repository)
        repository_dependency_tuple = (tool_shed_url, name, owner, changeset_revision)
        filepath = self.generate_temp_path("1430_repository_dependency")
        self.create_repository_dependency(column_repository, [repository_dependency_tuple], filepath=filepath)

    def test_0020_install_column_repository(self):
        """Install the column_1430 repository into Galaxy.

        This is step 1 (galaxy side) - Install the column_1430 repository, making sure to check the checkbox to
        handle repository dependencies so that the filter_1430 repository is also installed. Make sure to install
        the repositories in a specified section of the tool panel.
        """
        self._install_repository(
            "column_1430",
            common.test_user_1_name,
            category_name,
            new_tool_panel_section_label="repair",
            install_tool_dependencies=False,
            install_repository_dependencies=True,
        )

    def test_0025_uninstall_filter_repository(self):
        """Uninstall the filter_1430 repository from Galaxy.

        This is step 2 - Uninstall the filter_1430 repository.
        """
        installed_repository = self._get_installed_repository_by_name_owner("filter_1430", common.test_user_1_name)
        self._uninstall_repository(installed_repository)
        self._assert_has_no_installed_repos_with_names("filter_1430")



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

category_name = "Test 0480 Tool dependency definition validation"
category_description = "Test script 0480 for validating tool dependency definitions."
repository_name = "package_invalid_tool_dependency_xml_1_0_0"
repository_description = "Contains a tool dependency definition that should return an error."
repository_long_description = "This repository is in the test suite 0480"

"""

1. Create a repository package_invalid_tool_dependency_xml_1_0_0
2. Upload a tool_dependencies.xml file to the repository with no <actions> tags around the <action> tags.
3. Verify error message is displayed.

"""


class TestDependencyDefinitionValidation(ShedTwillTestCase):
    """Test the tool shed's tool dependency XML validation."""

    def test_0000_initiate_users_and_category(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.admin_email, username=common.admin_username)
        self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)

    def test_0005_create_tool_dependency_repository(self):
        """Create and populate package_invalid_tool_dependency_xml_1_0_0.

        This is step 1 - Create a repository package_invalid_tool_dependency_xml_1_0_0.

        Create a repository named package_invalid_tool_dependency_xml_1_0_0 that will contain only a single file named tool_dependencies.xml.
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.user_populator().setup_test_data_repo("0480", repository, assert_ok=False)

    def test_0010_populate_tool_dependency_repository(self):
        """Verify package_invalid_tool_dependency_xml_1_0_0.

        This is step 3 - Verify repository. The uploaded tool dependency XML should not have resulted in a new changeset.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        assert self.repository_is_new(
            repository
        ), "Uploading an incorrectly defined tool_dependencies.xml resulted in a changeset being generated."



"""
API operations allowing clients to determine Tool Shed instance's
capabilities and configuration settings.
"""

import logging

from galaxy.web import expose_api_anonymous_and_sessionless
from . import BaseShedAPIController

log = logging.getLogger(__name__)


class ConfigurationController(BaseShedAPIController):
    @expose_api_anonymous_and_sessionless
    def version(self, trans, **kwds):
        """
        GET /api/version
        Return a description of the version_major and version of Galaxy Tool Shed
        (e.g. 15.07 and 15.07.dev).

        :rtype:     dict
        :returns:   dictionary with versions keyed as 'version_major' and 'version'
        """
        return {"version_major": self.app.config.version_major, "version": self.app.config.version}
